#!/usr/local/bin/ruby
#
# A single "chunk" in a diff; may not be an appropriate abstraction.
# OTOH, it may well be a very good one (at least if integrated into a system
# that dynamically creates these as appropriate...)
#
class DiffChunk
	def initialize(inStart, inLineCount, outStart, outLineCount)
		@inStart  = inStart
		@inCount  = inLineCount
		@outStart = outStart
		@outCount = outLineCount
		@lineNum  = @inStart
		@delete   = {}
		@lines    = []
		@outLines = {}
	end
	def addInLine(line)
		@lines.push line
		@delete[@lineNum] = 1;
		@lineNum += 1
	end
	def addOutLine(line)
		@outLines[@lineNum] = [] unless @outLines.has_key? @lineNum;
		@outLines[@lineNum].push line
	end
	def addLine(line)
		@lines.push line
		@lineNum += 1
	end
	def printUnified(file)
		file.print "@@ -#@inStart,#@inCount +#@outStart,#@outCount @@\n";
		lineIn = @inStart;
		lineOut = @outStart
		@lines.each { |line|
			if (@delete[lineIn])
				file.print "-#{line}\n";
				lineIn += 1;
			else
				if @outLines[lineIn]
					@outLines[lineIn].each { |oline| file.print "+#{oline}\n" }
					lineOut += @outLines[lineIn].length
				end
				file.print " #{line}\n";
				lineIn += 1;
				lineOut += 1;
			end
		}
	end
	#
	# Check if this chunk fit with a particular set of lines (a file)
	#
	def checkFit(lines)
		i = -1
		@lines.each { |line|
			#print "\"#{line}\" != \"#{lines[@inStart + i]}\"\n" unless line == lines[@inStart + i]
			return false unless line == lines[@inStart + i]
			i += 1
		}
		return true
	end
end

#
# The diffs covering a single file
#
class FileDiff
  # Index: @index
  # RCS File: @rcsfile
  # Indate
  # Outdate
  # Format (unified, Context, old-fashioned, etc)
  # Free context lines
  # Added lines
  # Removed lines
  # Source file (?)
  # Destination file (?)

  # Start of class methods
  def initialize
  	@chunks = [];
  end
  #
  # Yield each chunk in the diff
  #
  def eachChunk
  	@chunks.each { |chunk| yield chunk }
  end
  #
  # Read a unified diff from a file
  #
  def readUnified(file)
	inLineCount = nil
	outLineCount = nil
	chunk = nil
	file.each do |line|
		if (inLineCount || outLineCount)
			if (line =~ /^\+(.*)/)
				chunk.addOutLine($1)
				outLineCount -= 1;
			elsif (line =~ /^-(.*)/)
				chunk.addInLine($1)
				inLineCount -= 1;
			elsif (line =~ /^ (.*)/)
				chunk.addLine($1);
				inLineCount  -= 1;
				outLineCount -= 1;
			else
				raise "Malformed diff: " + line;
			end
			inLineCount = nil if inLineCount == 0;
			outLineCount = nil if outLineCount == 0;
			chunk = nil unless inLineCount || outLineCount;
		elsif (line =~ /^Index:\s+(.*)/)
			@index = $1;
		elsif (line =~ /^RCS\s+file:\s+(.*)/)
			@rcsfile = $1;
		elsif (line =~ /^={67}$/)
			# Lose indicator line
		elsif (line =~ /^(diff\s+.*)/)
			@diffline = $1;
		elsif (line =~ /^(---\s+.*)/)
			@startline = $1;
		elsif (line =~ /^(\+\+\+\s+.*)/)
			@endline = $1;
		elsif (line =~ /^@@\s+\-(\d+),(\d+)\s+\+(\d+),(\d+)\s+@@(.*)/)
			# OK, we have got a new chunk.  Start processing it,
			# first of all by making a new chunk object...
			inStart = $1.to_i
			inLineCount = $2.to_i
			outStart = $3.to_i
			outLineCount = $4.to_i
			chunk = DiffChunk.new(inStart, inLineCount, outStart, outLineCount);
			@chunks.push(chunk);
		else
			#print "Skipped: " + line;
		end
	end
  end
  def printUnified(file)
  	file.print "Index: #@index\n#{"="*67}\n" if @index;
  	file.print "RCS file: #@rcsfile\n" if @rcsfile;
	file.print "#@diffline\n" if @diffline;
	file.print "#@startline\n" if @startline;
	file.print "#@endline\n" if @endline;
	eachChunk { |chunk| chunk.printUnified(file) }
  end
  #
  # Read in the source file for this diff.
  # This makes the FileDiff "complete", and allows us to do a host of new and
  # interesting operations on it.
  #
  def readSource(file)
  	lines = []
	file.each { |line| lines.push line.chomp; }
	@chunks.each { |chunk| raise "Source does not fit patch" unless chunk.checkFit(lines) }
  	@lines = lines
  end
end

diff = FileDiff.new;
#diff.readUnified(ARGF);
#diff.printUnified($stdout);
diff.readUnified(File.open("test.diff"));
diff.readSource(File.open("test"));
